@@ -51,11 +51,17 @@ use Clone 'clone';
 use JSON;
 use DBI;
 use Geo::GDAL;
+use HTTP::Date;
+use File::MkTemp;
 
 use Data::Dumper;
 use XML::LibXML::PrettyPrint;
 
-our $VERSION = '0.07';
+use Geo::OGC::Service;
+use vars qw(@ISA);
+push @ISA, qw(Geo::OGC::Service::Common);
+
+our $VERSION = '0.08';
 
 our %gml_geometry_type = (
     Envelope => 1,
@@ -271,17 +277,7 @@ sub GetCapabilities {
     $ns{version} = $self->{version};
 
     $writer->open_element('wfs:WFS_Capabilities', \%ns);
-    $writer->element('ows:ServiceIdentification', 
-                     [['ows:Title', $self->{config}{Title}],
-                      ['ows:Abstract'],
-                      ['ows:ServiceType', {codeSpace=>"OGC"}, 'OGC WFS'],
-                      ['ows:ServiceTypeVersion', $self->{config}{ServiceTypeVersion}],
-                      ['ows:Fees'],
-                      ['ows:AccessConstraints']]);
-    $writer->element('ows:ServiceProvider',
-                     [['ows:ProviderName'],
-                      ['ows:ProviderSite', {'xlink:type'=>"simple", 'xlink:href'=>""}],
-                      ['ows:ServiceContact']]);
+    $self->DescribeService($writer);
     $self->OperationsMetadata($writer);
     $self->FeatureTypeList($writer);
     $self->Filter_Capabilities($writer);
@@ -293,15 +289,19 @@ sub OperationsMetadata  {
     my ($self, $writer) = @_;
     $writer->open_element('ows:OperationsMetadata');
     my @versions = split /,/, $self->{config}{AcceptVersions} // '2.0.0,1.1.0,1.0.0';
-    $self->Operation($writer, 'GetCapabilities', 
+    $self->Operation($writer, 'GetCapabilities',
+                     { Get => 1, Post => 1 },
                      [{service => ['WFS']}, 
                       {AcceptVersions => \@versions}, 
                       {AcceptFormats => ['text/xml']}]);
     $self->Operation($writer, 'DescribeFeatureType', 
+                     { Get => 1, Post => 1 },
                      [{outputFormat => [sort keys %OutputFormats]}]);
     $self->Operation($writer, 'GetFeature',
+                     { Get => 1, Post => 1 },
                      [{resultType => ['results']}, {outputFormat => [sort keys %OutputFormats]}]);
     $self->Operation($writer, 'Transaction',
+                     { Get => 1, Post => 1 },
                      [{inputFormat => ['text/xml; subtype=gml/3.1.1']}, 
                       {idgen => ['GenerateNew','UseExisting','ReplaceDuplicate']},
                       {releaseAction => ['ALL','SOME']}
@@ -334,29 +334,6 @@ sub OperationsMetadata  {
     $writer->close_element;
 }
 
-sub Operation {
-    my($self, $writer, $name, $parameters) = @_;
-    my @parameters;
-    for my $p (@$parameters) {
-        for my $n (keys %$p) {
-            my @values;
-            for my $v (@{$p->{$n}}) {
-                push @values, ['ows:Value', $v];
-            }
-            push @parameters, ['ows:Parameter', {name=>$n}, \@values];
-        }
-    }
-    $writer->element(
-        'ows:Operation', 
-        {name => $name}, 
-        [['ows:DCP', 
-          ['ows:HTTP', 
-           [['ows:Get', {'xlink:type'=>'simple', 'xlink:href'=>$self->{config}{resource}}],
-            ['ows:Post', {'xlink:type'=>'simple', 'xlink:href'=>$self->{config}{resource}}]]
-          ]],
-         @parameters]);
-}
-
 sub FeatureTypeList  {
     my ($self, $writer) = @_;
     $writer->open_element('wfs:FeatureTypeList');
@@ -728,41 +705,108 @@ sub GetFeature {
     $format = $OutputFormats{$format} if $OutputFormats{$format};
     print STDERR "Output format (for GDAL creation): $format\n" if $self->{debug} > 2;
 
-    my $vsi = '/vsistdout/';
-    my $gml;
-    my $l2;
-
-    my $stdout = capture_stdout {
-        $gml = Geo::OGR::Driver('GML')->Create($vsi, { TARGET_NAMESPACE => $ns, PREFIX => $prefix, FORMAT => $format });
-        $l2 = $gml->CreateLayer($type->{Name});
+    my $i = 0;
+    my $count = $query->{count};
+    if (Geo::GDAL::VersionInfo >= 2010000) {
+        # use streaming object
+        #say STDERR "use streaming object";
+        my $content_type = $self->{config}{'Content-Type'} // 'text/xml; charset=utf-8';
+        my $writer = $self->{responder}->([200, [ 'Content-Type' => $content_type ]]);
+        my $gml = Geo::OGR::Driver('GML')->Create(
+            $writer, { TARGET_NAMESPACE => $ns, PREFIX => $prefix, FORMAT => $format });
+        my $l2 = $gml->CreateLayer($type->{Name});
         my $d = $layer->GetLayerDefn;
         for (0..$d->GetFieldCount-1) {
             my $f = $d->GetFieldDefn($_);
             $l2->CreateField($f);
         }
-    };
+        $layer->ResetReading;
+        while (my $f = $layer->GetNextFeature) {
+            $l2->CreateFeature($f);
+            $i++;
+            last if defined $count and $i >= $count;
+        }
+    }
+    elsif (Geo::GDAL::VersionInfo >= 2000200) {
+        # capture stdout, requires fix to close vsistdout bug
+        #say STDERR "capture stdout";
+        my $content_type = $self->{config}{'Content-Type'};
+        my $writer = Geo::OGC::Service::XMLWriter::Streaming->new($self->{responder}, $content_type);
 
-    my $content_type = $self->{config}{'Content-Type'} // 'text/xml';
-    my $writer = Geo::OGC::Service::XMLWriter::Streaming->new($self->{responder}, $content_type);
-    $writer->write($stdout);
+        my $vsi = '/vsistdout/';
+        my $gml;
+        my $l2;
 
-    my $count = $query->{count};
-    my $i = 0;
-    $layer->ResetReading;
-    while (1) {
-        my $f = $layer->GetNextFeature;
         my $stdout = capture_stdout {
-            if ($f) {
+            $gml = Geo::OGR::Driver('GML')->Create(
+                $vsi, { TARGET_NAMESPACE => $ns, PREFIX => $prefix, FORMAT => $format });
+            $l2 = $gml->CreateLayer($type->{Name});
+            my $d = $layer->GetLayerDefn;
+            for (0..$d->GetFieldCount-1) {
+                my $f = $d->GetFieldDefn($_);
+                $l2->CreateField($f);
+            }
+        };        
+        $writer->write($stdout);
+        
+        $layer->ResetReading;
+        while (1) {
+            my $f = $layer->GetNextFeature;
+            my $stdout = capture_stdout {
+                if ($f) {
+                    $l2->CreateFeature($f);
+                    $i++;
+                }
+                if (!$f or (defined $count and $i >= $count)) {
+                    undef $l2;
+                    undef $gml;
+                }
+            };
+            $writer->write($stdout);
+            last unless $gml;
+        }
+    }
+    else {
+        # use temp file
+        #say STDERR "use tmp file";
+        my $file = '/tmp/' . mktemp('tempXXXXXX', '/tmp').'.xml';
+        my $content_type = $self->{config}{'Content-Type'} // 'text/xml; charset=utf-8';
+        {
+            # local variables, to close the file at the end
+            my $gml = Geo::OGR::Driver('GML')->Create(
+                $file, { TARGET_NAMESPACE => $ns, PREFIX => $prefix, FORMAT => $format });
+            my $l2 = $gml->CreateLayer($type->{Name});
+            my $d = $layer->GetLayerDefn;
+            for (0..$d->GetFieldCount-1) {
+                my $f = $d->GetFieldDefn($_);
+                $l2->CreateField($f);
+            }
+            
+            $layer->ResetReading;
+            while (my $f = $layer->GetNextFeature) {
                 $l2->CreateFeature($f);
                 $i++;
+                last if defined $count and $i >= $count;
             }
-            if (!$f or (defined $count and $i >= $count)) {
-                undef $l2;
-                undef $gml;
-            }
-        };
-        $writer->write($stdout);
-        last unless $gml;
+        }
+
+        open my $fh, "<:raw", $file or return $self->return_403;
+
+        my @stat = stat $file;
+    
+        Plack::Util::set_io_path($fh, Cwd::realpath($file));
+    
+        $self->{responder}->([ 200, 
+                               [
+                                'Content-Type'   => $content_type,
+                                'Content-Length' => $stat[7],
+                                'Last-Modified'  => HTTP::Date::time2str( $stat[9] )
+                               ],
+                               $fh,
+                             ]);
+        unlink $file;
+        $file =~ s/xml$/xsd/;
+        unlink $file;
     }
     print STDERR "$i features served, max is ",$count//'not set',"\n" if $self->{debug};
 }
